import anthropic
from pandalytics.tools import tool, pandas_executor
from dotenv import load_dotenv
import pandas as pd
from io import StringIO
from typing import List

from pandalytics.templates.agent_prompt import (
    PREFIX,
    FORMAT_INSTRUCTIONS,
    SUFFIX,
    TEMPLATE_TOOL_RESPONSE,
)


class Pandalytics:
    """
    Main agent class
    """

    def __init__(self, file: List[str]) -> None:
        ## csv file
        self.file = file

        ## to be implemented
        self.pandas_code = None
        self.code_result = None

    ## private function to extract the dataframe context
    def _extract_data_context(self):
        df = pd.read_csv(self.file)

        info_output = StringIO()
        df.info(buf=info_output)

        # Get the string value
        info_output_string = info_output.getvalue()

        context = (
            "The columns of the dataframe are: {}\n\n"
            "The overall information of the dataframe is:\n{}\n\n"
            "The unique values of the columns are:\n{}"
        ).format(df.columns, info_output_string, df.apply(pd.Series.unique))
        return context

    def query(self, query: str):
        """
        Method to execute the the user query

        Args:
            query (str): A user query
        """
        load_dotenv()

        metadata = self._extract_data_context()

        client = anthropic.Anthropic()
        response = client.beta.tools.messages.create(
            model="claude-3-haiku-20240307",
            temperature=0,
            max_tokens=1024,
            system=PREFIX.format(tool=tool) + "\n" + FORMAT_INSTRUCTIONS,
            tools=tool,
            messages=[
                {
                    "role": "user",
                    "content": SUFFIX.format(
                        query=query, file=self.file, metadata=metadata
                    ),
                }
            ],
        )

        if response.stop_reason == "tool_use":
            tool_use = next(
                block for block in response.content if block.type == "tool_use"
            )
            tool_name = tool_use.name
            tool_input = tool_use.input

        if tool_name == "pandas_executor":
            out = pandas_executor(
                tool_input["file"], tool_input["query"], tool_input["metadata"]
            )
            tool_result = out[1]
            self.pandas_code = out[0]
            self.code_result = tool_result

        result = client.beta.tools.messages.create(
            model="claude-3-haiku-20240307",
            temperature=0,
            max_tokens=1024,
            messages=[
                {
                    "role": "user",
                    "content": SUFFIX.format(
                        query=query, file=self.file, metadata=metadata
                    ),
                },
                {"role": "assistant", "content": response.content},
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "tool_result",
                            "tool_use_id": tool_use.id,
                            "content": TEMPLATE_TOOL_RESPONSE.format(
                                observation=tool_result
                            ),
                        }
                    ],
                },
            ],
            tools=tool,
        )

        return next(
            (block.text for block in result.content if hasattr(block, "text")),
            None,
        )



PREFIX = """As a world-class data analyst proficient in writing pandas code, you specialize in retrieving data from dataframes to derive insightful conclusions and create compelling visualizations using matplotlib graphs.
Meticulously considering dataframe metadata, including column names, data types, and unique values, You ensure precision in code composition, thus minimizing errors when filtering and comparing data. 
Your reliability is unmatched, consistently delivering accurate solutions."""

FORMAT_INSTRUCTIONS = """RESPONSE FORMAT INSTRUCTIONS
----------------------------
When you generate the pandas code, Update the initial code:

```python
# TODO: import the required dependencies
import matplotlib.pyplot as plt
import pandas as pd

## declare result variable here

# Write code here

print(result)
```

Generate python code and return full updated code:
"""

SUFFIX = """
Take a deep breath and thoroughly analyze the dataframe's metadata to gain a comprehensive understanding. The metadata is as follows:\n {metadata}
Next, carefully analyze the user query: {query} and extract all necessary data from the metadata to generate the required pandas code.
The file path is: {file_path}

You needs to pay special attention while writing code for comparing the values. Use the exact values from csv file for comparing the values.

Ensure that the generated code can run independently without any additional modifications.
**Note:** You MUST only provide the pandas code as output.
"""



import subprocess
import threading
import time


class PythonShell:
    def __init__(self):
        self.process = None
        self.output = ""
        self.error = ""

    def execute_code(self, code):
        try:
            self.process = subprocess.Popen(
                ["python", "-c", code],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                shell=False,
            )
            self._capture_output()
            self.process.wait()
        except Exception as e:
            print(f"Error executing command: {e}")

    def _capture_output(self):
        self.output, self.error = self.process.communicate()

    def get_output(self):
        return self.output

    def get_error(self):
        return self.error



PREFIX = """You are Pandalytics Agent, an advanced analytical tool powered by pandas, meticulously crafted by Team Zero to extract insights and answer user queries efficiently from tabular data.

Equipped with an arsenal of cutting-edge tools, Pandalytics Agent seamlessly generates and executes pandas code, providing comprehensive solutions to diverse data challenges. Leveraging dynamic code generation, precise execution, and compelling visualizations, it transforms raw data into actionable insights.

Embark on a journey of exploration and discovery with Pandalytics Agent as your trusted companion, navigating the complexities of data analysis effortlessly.

TOOLS
------
Unlock the full potential of Pandalytics Agent with direct access to the following tools: {tool}
"""

FORMAT_INSTRUCTIONS = """RESPONSE FORMAT INSTRUCTIONS
----------------------------
When you use tools or generate final answer, please output a response in the given formats:
**Explain and provide analysis**
If the response involves providing the insight and using a tool, you can start using the tool and then generate precise to the point answer. 
"""


SUFFIX = """The user query is: {query}\nThe file path is {file}.\nThe metadata of the dataframe is {metadata}"""


TEMPLATE_TOOL_RESPONSE = """TOOL RESPONSE:
---------------------
{observation}

THOUGHT
--------------------

Let's assess the tool response to answer the human's initial query. Please follow these instructions:
- After assessing the response of tool, provide to the point answer to human query.
- MUST NOT include explanations of the tool's functions.
- MUST answer based on the output of tool. If there is no output from tool, MUST reply 'Information does not present in data'
"""



from dotenv import load_dotenv
import anthropic

from pandalytics.shell import PythonShell
from  pandalytics.templates.code_generation_prompt import PREFIX, FORMAT_INSTRUCTIONS, SUFFIX


## abstraction of anthropic
## for re-usability
def anthropic_client(system_prompt: str, user_prompt: str):
    load_dotenv()
    client = anthropic.Anthropic()
    message = client.messages.create(
        model="claude-3-haiku-20240307",
        temperature=0,
        system=system_prompt,
        max_tokens=1024,
        messages=[{"role": "user", "content": user_prompt}],
    )
    return next(
        (block.text for block in message.content if hasattr(block, "text")),
        None,
    )


## function for tool
def pandas_executor(file_path: str, query: str, metadata: str):
    system_prompt = PREFIX + "\n" + FORMAT_INSTRUCTIONS
    user_prompt = SUFFIX.format(metadata=metadata, query=query, file_path=file_path)

    pandas_code = anthropic_client(system_prompt=system_prompt, user_prompt=user_prompt)
    code = pandas_code.strip("`")[6:]

    pandas_shell = PythonShell()
    pandas_shell.execute_code(code=code)

    if pandas_shell.get_error() != "":
        return code, pandas_shell.get_error()

    return code, pandas_shell.get_output()


tool = [
    {
        "name": "pandas_executor",
        "description": "Generate and execute the pandas code",
        "input_schema": {
            "type": "object",
            "properties": {
                "file": {
                    "type": "string",
                    "description": "The file path of csv file containing the data, e.g. records.csv, movies.csv, ratings.csv",
                },
                "query": {
                    "type": "string",
                    "description": "The user query to get the code, e.g. Give me the top 5 countries by sales?",
                },
                "metadata": {
                    "type": "string",
                    "description": "The metadata about the pandas dataframe.",
                },
            },
            "required": ["file_path", "query", "metadata"],
        },
    }
]
